@@ -14,14 +14,13 @@ use buzz::{s3, CachedFile, RangeCache};
 
 enum Source {
     File(String),
-    S3(String),
+    S3(String, usize),
 }
 
-fn output_rows<T: 'static>(reader: T, offset: u32, limit: i32)
+fn output_rows<T: 'static>(file_reader: SerializedFileReader<T>, offset: u32, limit: i32)
 where
     T: ChunkReader,
 {
-    let file_reader = SerializedFileReader::new(reader).unwrap();
     let iter = file_reader.get_row_iter(None).unwrap();
 
     let mut input_rows_count = 0;
@@ -45,9 +44,10 @@ async fn print_json_from(source: Source, offset: u32, limit: i32) {
     match source {
         Source::File(path) => {
             let file = File::open(&Path::new(&path)).unwrap();
-            output_rows(file, offset, limit);
+            let file_reader = SerializedFileReader::new(file).unwrap();
+            output_rows(file_reader, offset, limit);
         }
-        Source::S3(url_str) => {
+        Source::S3(url_str, concurrency) => {
             let url = Url::parse(&url_str).unwrap();
             let host_str = url.host_str().unwrap();
             let key = &url.path()[1..];
@@ -62,13 +62,28 @@ async fn print_json_from(source: Source, offset: u32, limit: i32) {
             let object_found = &list_res.contents.unwrap()[0];
             let size = object_found.size.unwrap().try_into().unwrap();
 
-            let cache = RangeCache::new().await;
+            let is_parallel = concurrency > 1;
+            let cache = RangeCache::new(concurrency).await;
             let (dler_id, dler_creator) = s3::downloader_creator(Region::default().name());
             let file_id = s3::file_id(host_str, key);
-            let file = CachedFile::new(file_id, size, Arc::new(cache), dler_id, dler_creator);
+            let file = CachedFile::new(
+                file_id,
+                size,
+                Arc::new(cache),
+                dler_id,
+                dler_creator,
+                is_parallel,
+            );
 
-            file.prefetch(file.len() - size, size as usize);
-            output_rows(file, offset, limit);
+            let pre_size: u64 = if is_parallel { 1024 * 1024 } else { size - 4 };
+            file.prefetch(file.len() - pre_size, pre_size as usize);
+
+            let file_reader = SerializedFileReader::new(file).unwrap();
+
+            let blocking_task = tokio::task::spawn_blocking(move || {
+                output_rows(file_reader, offset, limit);
+            });
+            blocking_task.await.unwrap();
         }
     };
 }
@@ -100,14 +115,23 @@ async fn main() {
                 .about("Maximum number of rows to output")
                 .takes_value(true),
         )
+        .arg(
+            Arg::new("concurrency")
+                .short('c')
+                .long("concurrency")
+                .value_name("NUMBER")
+                .about("Maximum number of concurrent column downloads (S3 only, default: 10)")
+                .takes_value(true),
+        )
         .get_matches();
 
     let offset: u32 = matches.value_of_t("offset").unwrap_or(0);
     let limit: i32 = matches.value_of_t("limit").unwrap_or(-1);
     let file: String = matches.value_of_t("FILE").unwrap_or_else(|e| e.exit());
+    let concurrency: usize = matches.value_of_t("concurrency").unwrap_or(10);
 
     if file.as_str().starts_with("s3://") {
-        print_json_from(Source::S3(file), offset, limit).await;
+        print_json_from(Source::S3(file, concurrency), offset, limit).await;
     } else {
         print_json_from(Source::File(file), offset, limit).await;
     }
